using System;
using System.Data;
using System.Linq;
using System.Linq.Dynamic.Core;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Text.Encodings.Web;
using Microsoft.AspNetCore.Components;
using Microsoft.Data.SqlClient;
using Microsoft.EntityFrameworkCore;
using Radzen;

using Razen2.Server.Data;

namespace Razen2.Server
{
    public partial class Radzen2Service
    {
        Radzen2Context Context
        {
           get
           {
             return this.context;
           }
        }

        private readonly Radzen2Context context;
        private readonly NavigationManager navigationManager;

        public Radzen2Service(Radzen2Context context, NavigationManager navigationManager)
        {
            this.context = context;
            this.navigationManager = navigationManager;
        }

        public void Reset() => Context.ChangeTracker.Entries().Where(e => e.Entity != null).ToList().ForEach(e => e.State = EntityState.Detached);


        public async Task ExportProjectsToExcel(Query query = null, string fileName = null)
        {
            navigationManager.NavigateTo(query != null ? query.ToUrl($"export/radzen2/projects/excel(fileName='{(!string.IsNullOrEmpty(fileName) ? UrlEncoder.Default.Encode(fileName) : "Export")}')") : $"export/radzen2/projects/excel(fileName='{(!string.IsNullOrEmpty(fileName) ? UrlEncoder.Default.Encode(fileName) : "Export")}')", true);
        }

        public async Task ExportProjectsToCSV(Query query = null, string fileName = null)
        {
            navigationManager.NavigateTo(query != null ? query.ToUrl($"export/radzen2/projects/csv(fileName='{(!string.IsNullOrEmpty(fileName) ? UrlEncoder.Default.Encode(fileName) : "Export")}')") : $"export/radzen2/projects/csv(fileName='{(!string.IsNullOrEmpty(fileName) ? UrlEncoder.Default.Encode(fileName) : "Export")}')", true);
        }

        partial void OnProjectsRead(ref IQueryable<Razen2.Server.Models.Radzen2.Project> items);

        public async Task<IQueryable<Razen2.Server.Models.Radzen2.Project>> GetProjects(Query query = null)
        {
            var items = Context.Projects.AsQueryable();


            if (query != null)
            {
                if (!string.IsNullOrEmpty(query.Expand))
                {
                    var propertiesToExpand = query.Expand.Split(',');
                    foreach(var p in propertiesToExpand)
                    {
                        items = items.Include(p.Trim());
                    }
                }

                if (!string.IsNullOrEmpty(query.Filter))
                {
                    if (query.FilterParameters != null)
                    {
                        items = items.Where(query.Filter, query.FilterParameters);
                    }
                    else
                    {
                        items = items.Where(query.Filter);
                    }
                }

                if (!string.IsNullOrEmpty(query.OrderBy))
                {
                    items = items.OrderBy(query.OrderBy);
                }

                if (query.Skip.HasValue)
                {
                    items = items.Skip(query.Skip.Value);
                }

                if (query.Top.HasValue)
                {
                    items = items.Take(query.Top.Value);
                }
            }

            OnProjectsRead(ref items);

            return await Task.FromResult(items);
        }

        partial void OnProjectGet(Razen2.Server.Models.Radzen2.Project item);

        public async Task<Razen2.Server.Models.Radzen2.Project> GetProjectById(Guid id)
        {
            var items = Context.Projects
                              .AsNoTracking()
                              .Where(i => i.Id == id);

  
            var itemToReturn = items.FirstOrDefault();

            OnProjectGet(itemToReturn);

            return await Task.FromResult(itemToReturn);
        }

        partial void OnProjectCreated(Razen2.Server.Models.Radzen2.Project item);
        partial void OnAfterProjectCreated(Razen2.Server.Models.Radzen2.Project item);

        public async Task<Razen2.Server.Models.Radzen2.Project> CreateProject(Razen2.Server.Models.Radzen2.Project project)
        {
            OnProjectCreated(project);

            var existingItem = Context.Projects
                              .Where(i => i.Id == project.Id)
                              .FirstOrDefault();

            if (existingItem != null)
            {
               throw new Exception("Item already available");
            }            

            try
            {
                Context.Projects.Add(project);
                Context.SaveChanges();
            }
            catch
            {
                Context.Entry(project).State = EntityState.Detached;
                throw;
            }

            OnAfterProjectCreated(project);

            return project;
        }

        public async Task<Razen2.Server.Models.Radzen2.Project> CancelProjectChanges(Razen2.Server.Models.Radzen2.Project item)
        {
            var entityToCancel = Context.Entry(item);
            if (entityToCancel.State == EntityState.Modified)
            {
              entityToCancel.CurrentValues.SetValues(entityToCancel.OriginalValues);
              entityToCancel.State = EntityState.Unchanged;
            }

            return item;
        }

        partial void OnProjectUpdated(Razen2.Server.Models.Radzen2.Project item);
        partial void OnAfterProjectUpdated(Razen2.Server.Models.Radzen2.Project item);

        public async Task<Razen2.Server.Models.Radzen2.Project> UpdateProject(Guid id, Razen2.Server.Models.Radzen2.Project project)
        {
            OnProjectUpdated(project);

            var itemToUpdate = Context.Projects
                              .Where(i => i.Id == project.Id)
                              .FirstOrDefault();

            if (itemToUpdate == null)
            {
               throw new Exception("Item no longer available");
            }
                
            var entryToUpdate = Context.Entry(itemToUpdate);
            entryToUpdate.CurrentValues.SetValues(project);
            entryToUpdate.State = EntityState.Modified;

            Context.SaveChanges();

            OnAfterProjectUpdated(project);

            return project;
        }

        partial void OnProjectDeleted(Razen2.Server.Models.Radzen2.Project item);
        partial void OnAfterProjectDeleted(Razen2.Server.Models.Radzen2.Project item);

        public async Task<Razen2.Server.Models.Radzen2.Project> DeleteProject(Guid id)
        {
            var itemToDelete = Context.Projects
                              .Where(i => i.Id == id)
                              .FirstOrDefault();

            if (itemToDelete == null)
            {
               throw new Exception("Item no longer available");
            }

            OnProjectDeleted(itemToDelete);


            Context.Projects.Remove(itemToDelete);

            try
            {
                Context.SaveChanges();
            }
            catch
            {
                Context.Entry(itemToDelete).State = EntityState.Unchanged;
                throw;
            }

            OnAfterProjectDeleted(itemToDelete);

            return itemToDelete;
        }
        }
}